@@ -3,6 +3,8 @@
 
 import re
 
+from .scim import SCIM
+
 
 class Update:
     """Slack Update class."""
@@ -21,9 +23,12 @@ class Update:
         if "verbose" in args or auth.verbose:
             self.verbose = True
 
+        self.emplids = None
+        self.github_profile_id = None
         self.google_users = None
         self.other_accounts = None
         self.people = None
+        self.scim_users = None
         self.slack_emails = None
         self.slack_users = None
         self.team_profile = None
@@ -44,17 +49,17 @@ class Update:
         # skip users with no email
         if not email:
             print(f"ERROR: User with no email: {real_name}")
-            return
+            return None
 
         # check for an existing other account
         if email in self.other_accounts:
-            return
+            return None
 
         # check for an existing person
         if email not in self.people:
             if not user["deleted"]:
                 print(f"ERROR: Unknown user: {real_name} <{email}>")
-            return
+            return None
 
         # get person
         person = self.people[email]
@@ -116,6 +121,23 @@ class Update:
                 return field["id"]
         return None
 
+    def _get_google_group_emails(self, email):
+        """Return the list of Slack users in a Google group."""
+        g = self.auth.google()
+        g.auth_service_account(g.scopes, g.subject)
+        members = g.directory().get_derived_members(email)
+        emails = []
+        for member in members:
+            if member["type"] != "USER":
+                continue
+            member_id = member["id"]
+            if member_id in self.google_users:
+                email = self.google_users[member_id]["primaryEmail"]
+            else:
+                email = member["email"].lower()
+            emails.append(email)
+        return emails
+
     def _get_google_group_users(self, email):
         """Return the list of Slack users in a Google group."""
         g = self.auth.google()
@@ -134,7 +156,8 @@ class Update:
                 continue
             user = self.slack_emails[email]
             user_id = user["id"]
-            if not user["deleted"] and user_id not in users:
+            # skip deleted and single-channel guest users
+            if not user["deleted"] and not user["is_ultra_restricted"] and user_id not in users:
                 users.append(user_id)
         return users
 
@@ -145,6 +168,16 @@ class Update:
         fields = "nextPageToken,users(id,primaryEmail,name/fullName,suspended)"
         return g.directory().get_users_dict(fields=fields)
 
+    def _get_manager_user_id(self, person):
+        """Return the manager value for scim."""
+        manager_user_id = None
+        manager_id = person["manager_id"]
+        if manager_id in self.emplids:
+            manager_email = self.emplids[manager_id]["email_username"]
+            if manager_email in self.scim_users:
+                manager_user_id = self.scim_users[manager_email]["id"]
+        return manager_user_id
+
     def _get_other_accounts(self):
         """Return a dict of other accounts from BITSdb."""
         mongo = self.auth.mongo()
@@ -166,8 +199,8 @@ class Update:
             if self.verbose:
                 print("Getting Slack users...")
             self.slack_users = self.slack.get_users_dict()
-        if self.verbose:
-            print(f"Found {len(self.slack_users)} Slack users.")
+            if self.verbose:
+                print(f"Found {len(self.slack_users)} Slack users.")
         emails = {}
         for user_id in self.slack_users:
             user = self.slack_users[user_id]
@@ -208,7 +241,10 @@ class Update:
         g = self.auth.google()
         firestore = g.firestore(project)
         people = {}
+        self.emplids = {}
         for person in firestore.get_collection(collection):
+            emplid = person["emplid"]
+            self.emplids[emplid] = person
             key = person["email_username"]
             if not key:
                 continue
@@ -226,6 +262,179 @@ class Update:
             suffix = re.sub(r".* (\(.*/.*\))$", r"\1", last_name).replace(" ", "")
         return suffix
 
+    def _prepare_idp_user(self, person):
+        """Prepare a user for the Slack IDP."""
+        user = {
+            "active": not person["terminated"],
+            "name": {
+                "familyName": person["last_name"],
+                "givenName": person["first_name"],
+                "honorificPrefix": None,
+            },
+            "phoneNumbers": [
+                {
+                    "type": "mobile",
+                    "value": None,
+                },
+                {
+                    "primary": True,
+                    "value": person["primary_work_phone"],
+                }
+            ],
+            "profileUrl": f"https://people-cloud.broadinstitute.org/people/{person['emplid']}",
+            "timezone": "US/New_York",
+            "title": person["title"],
+            "urn:scim:schemas:extension:enterprise:1.0": {
+                "costCenter": None,
+                "department": person["department_name"],
+                "division": person["org_unit"] or None,
+                "employeeNumber": person["emplid"],
+                "organization": person["home_institution"],
+            },
+            "userName": person["username"],
+            "userType": person["worker_type"],
+        }
+
+        manager_id = self._get_manager_user_id(person)
+        if manager_id:
+            user["urn:scim:schemas:extension:enterprise:1.0"]["manager"] = {
+                "managerId": manager_id
+            }
+
+        return user
+
+    def idp_groups(self):
+        """Update the IDP Groups in Slack Enterprise Grid."""
+        scim = SCIM(self.slack.token)
+
+        if self.verbose:
+            print("Getting Slack Usergroups Syncs from BITSdb Mongo DB...")
+        syncs = self._get_slack_usergroup_syncs()
+        if self.verbose:
+            print(f"Usergroup Sync: {len(syncs)}")
+
+        if self.verbose:
+            print("Getting Slack Groups from SCIM API...")
+        groups = scim.get_groups_dict()
+        if self.verbose:
+            print(f"Groups: {len(groups)}")
+
+        if self.verbose:
+            print("Getting Slack Users from SCIM API...")
+        users = scim.get_users_dict(key="email")
+        if self.verbose:
+            print(f"Users: {len(users)}")
+
+        self.google_users = self._get_google_users()
+
+        sync_groups = []
+
+        # create missing groups and update groups
+        for sync in sorted(syncs, key=lambda x: x["google_group"]):
+
+            # get google group email
+            google_group_email = sync["google_group"]
+            if not google_group_email.endswith("@broadinstitute.org"):
+                print(f"Skipping non-Broad group: {google_group_email}")
+                continue
+
+            name = google_group_email.replace("@broadinstitute.org", "")
+
+            if name in sync_groups:
+                print(f"Duplicate sync group: {name}")
+                continue
+
+            # get google group members
+            google_group_members = self._get_google_group_emails(google_group_email)
+
+            if self.verbose:
+                print(f"Syncing group: {google_group_email}")
+
+            # assemble group member emails into slack users
+            google_group_slack_user_ids = []
+            for email in google_group_members:
+                if email in users:
+                    google_group_slack_user_ids.append(users[email]["id"])
+            google_group_slack_user_ids = sorted(set(google_group_slack_user_ids))
+
+            # assemble group users into group members
+            members = scim.prepare_group_members(google_group_slack_user_ids)
+
+            if name not in groups:
+                print(f"Creating group: {name}")
+                print(scim.create_group(name=name, members=members))
+
+            else:
+                group = groups[name]
+                group_member_ids = scim.get_group_member_ids(group)
+                if group_member_ids != google_group_slack_user_ids:
+                    print(f"Updating group from {len(group_member_ids)} to {len(google_group_slack_user_ids)} members: {name}")
+                    group["members"] = members
+                    scim.update_group(group["id"], group)
+
+            sync_groups.append(name)
+
+        # find groups to delete
+        for name in groups:
+            group = groups[name]
+            if name not in sync_groups:
+                print(f"Deleting group: {name}")
+                scim.delete_group(group.id)
+
+    def idp_users(self):
+        """Update IDP users in Slack."""
+        scim = SCIM(self.slack.token)
+
+        if self.verbose:
+            print("Getting Slack Users from SCIM API...")
+        self.scim_users = scim.get_users_dict(key="email")
+        if self.verbose:
+            print(f"Slack Users: {len(self.scim_users)}")
+
+        if self.verbose:
+            print("Getting People from Firestore...")
+        self.people = self._get_people()
+        if self.verbose:
+            print(f"People: {len(self.people)}")
+
+        for email in self.scim_users:
+            user = self.scim_users[email]
+
+            # fix manager data
+            if "urn:scim:schemas:extension:enterprise:1.0" in user:
+                if "manager" in user["urn:scim:schemas:extension:enterprise:1.0"]:
+                    if not user["urn:scim:schemas:extension:enterprise:1.0"]["manager"].get("managerId"):
+                        del user["urn:scim:schemas:extension:enterprise:1.0"]["manager"]
+
+            # skip non-people
+            if email not in self.people:
+                if email.endswith("@broadinstitute.org"):
+                    if self.verbose:
+                        print(f"Skipping unknown person: {email}")
+                continue
+
+            person = self.people[email]
+            idp_user = self._prepare_idp_user(person)
+
+            updates = []
+            patch = {}
+
+            for key in sorted(idp_user):
+                new = idp_user[key]
+                old = user.get(key)
+                if new != old:
+                    updates.append(f"  {key}: {old} -> {new}")
+                    patch[key] = new
+                    user[key] = new
+
+            if updates:
+                print(f"\nUpdating user: {email}...")
+                print("\n".join(updates))
+                try:
+                    scim.update_user(user["id"], user)
+                except Exception as e:
+                    print(e)
+
     def profiles(self):
         """Update user profiles in Slack."""
         if not self.team_profile:
@@ -384,10 +593,6 @@ class Update:
         for user_id in sorted(self.slack_users, key=lambda x: self.slack_users[x]["profile"].get("email", "")):
             user = self.slack_users[user_id]
 
-            # skip deleted users
-            # if user["deleted"]:
-            #     continue
-
             # count and skip bot users
             if user["is_bot"] or user["id"] == "USLACKBOT":
                 bots.append(user)
